package org.coows;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.coows.engine.CooEngine;
import org.coows.engine.CooException;

import org.coows.util.CooAgentState;
import org.coows.util.CooRequestID;
import org.coows.util.CooSetOfIntention;
import org.coows.util.CooSetOfPlans;


import java.util.Calendar;
import java.util.Vector;

/**
 * The CooWS agent
 */
public class CooAgent
    extends Thread {
    /**
     * Logger for this class
     */
    private static final Log logger = LogFactory.getLog(CooAgent.class);

    /* Retrieval strategy constants */
    public static final int RETRIEVAL_STRATEGY_ALWAYS = 0;

    public static final int RETRIEVAL_STRATEGY_NOLOCAL = 1;

    /* Acquisition strategy constants */
    public static final int ACQUISITION_STRATEGY_ADD = 0;

    public static final int ACQUISITION_STRATEGY_REPLACE = 1;

    public static final int ACQUISITION_STRATEGY_DISCARD = 2;

    /**
     * The agent ID
     */
    private String mAgentID;

    /**
     * The list of known agents
     */
    private Vector mPartners;

    /**
     * The agent plan library
     * 
     * @uml.property name="mPlans"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private CooSetOfPlans mPlans;


    /**
     * The acquisition strategy.<br>
     * Values range within the following values: <br>
     * CooAgent.ACQUSITION_STRATEGY_ADD <br>
     * CooAgent.ACQUISITION_STRATEGY_REPLACE <br>
     * CooAgent.ACQUISITION_STRATEGY_DISCARD
     */
    private int mAcquisitionStrategy;

    /**
     * The retrieval strategy. <br>
     * Values range within the following values: <br>
     * CooAgent.RETRIEVAL_STRATEGY_ALWAYS <br>
     * CooAgent.RETRIEVAL_STRATEGY_NOLOCAL
     */
    private int mRetrievalStrategy;

    /**
     * The intentions table <br>
     * For each intention ID a CooIntention is
     * 
     * @uml.property name="mIntentions"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private CooSetOfIntention mIntentions;

    /**
     * The agent's event queue
     * 
     * @uml.property name="mEventQueue"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private CooAgentEventsQueue mEventQueue;


    private boolean mPrepare4shutdown;

    private boolean mShutdown;

    /**
     * Constructor
     * 
     * @param anAgentState
     *            A previously serialized agent state
     * @throws CooException
     */
    public CooAgent(CooAgentState anAgentState)
        throws CooException {
        init();
        mAgentID = anAgentState.getMAgentID();
        mPlans = anAgentState.getMSop();
        mPartners = anAgentState.getMPartners();
        mAcquisitionStrategy = anAgentState.getMAcquisitionStrategy();
        mRetrievalStrategy = anAgentState.getMRetrievalStrategy();

        if (logger.isDebugEnabled()) {
            logger.debug("CooAgent() - agent " + mAgentID + " created");
        }
    }

    /**
     * Auxiliary constructor (for debug and test)
     * 
     * @param anAgentID
     * @param anAcquisitionSTRATEGY
     * @param aSop
     * @throws CooException
     */
    public CooAgent(
            String anAgentID, int anAcquisitionStrategy, CooSetOfPlans aSop)
        throws CooException {
        init();

        mAgentID = anAgentID;
        mAcquisitionStrategy = anAcquisitionStrategy;
        mPlans = aSop;

        if (logger.isDebugEnabled()) {
            logger.debug("CooAgent() - agent " + mAgentID + " created");
        }
    }


    

    /**
     * Init the local data structures (event queue, workers pool and boolean
     * variables)
     * 
     * @throws CooException
     */
    public void init()
            throws CooException {
        //mEventQueueOLD = new LinkedList();
        mEventQueue = new CooAgentEventsQueue(this);
        //mWorkers = new GenericObjectPool(new CooWorkerProcessEventFactory());
        //mMaxActiveWorkers = Integer.parseInt(CooEngine.getMConfig().getText(
          //      "/config/pools/agentWorkers/maxActive"));
        //mWorkers.setMaxActive(mMaxActiveWorkers);
        mPrepare4shutdown = false;
        mShutdown = false;
    }

    /**
     * Process new incoming events (assigns the work to a new thread or enqueue
     * locally)
     * 
     * @param anEvent
     */
    public void onEvent(ICooEvent anEvent) {
        if (logger.isDebugEnabled()) {
            logger.debug("agent: " + mAgentID + ", " + "onEvent(anEvent = "
                    + anEvent + ")");
        }

        // if we are going to shutdown
        if (true == mPrepare4shutdown)
        // discharge all events
            return;

        mEventQueue.enqueue(anEvent);
    }

    /**
     * Callback for "achieve" requests
     * 
     * @param anEvent
     */
    public void onEventAchieve(CooEventAchieve anEvent) {
        if (logger.isDebugEnabled()) {
            logger.debug("Agent: " + mAgentID 
                    + "onEventAchieve(" + anEvent + ")");
        }

        CooSetOfPlans relPlans = mPlans.getRelevant(anEvent.getMDes());

        CooInstanceRecord instRec =
            new CooInstanceRecord(anEvent.getMDes(), relPlans);

        
        try {
            CooIntention theInt;
            theInt = mIntentions.getIntention(anEvent.getMIntentionID());
            String instID = theInt.push(instRec);
            // TODO lanciare in esecuzione un piano se relplans NON vuoto

            if ((RETRIEVAL_STRATEGY_ALWAYS == mRetrievalStrategy)
                    || ((RETRIEVAL_STRATEGY_NOLOCAL == mRetrievalStrategy)
                            && relPlans.isEmpty()))
                sendRequested(
                        anEvent.getMDes(), anEvent.getMIntentionID(), instID);
        }
        catch (CooException e) {
            if (logger.isErrorEnabled()) {
                logger.error("Agent "+ mAgentID +" encountered an exception " +
                        "processing an <achieve> message for intention " + 
                        anEvent.getMIntentionID(),e);
            }
        }

    }

    public void onEventOrdinary(CooEventOrdinary anEvent) {
        String intID =
            String.valueOf(Calendar.getInstance().getTimeInMillis());

        // Create a new intention
        CooIntention newInt = new CooIntention(mAgentID, intID, 
                anEvent.getMSuccessAction(), anEvent.getMFailureAction());

        
        // Retrieve relevant plans
        CooSetOfPlans relPlans = mPlans.getRelevant(anEvent.getMDesire());

        // Create the first instance record for the intention stack
        CooInstanceRecord instRec =
            new CooInstanceRecord(anEvent.getMDesire(), relPlans);

        // Push the record on the stack
        newInt.push(instRec);

        // Update the intentions set 
        mIntentions.addIntention(newInt);

        // If available, run the first relevant plan
        if (!relPlans.isEmpty())
            instRec.runNextPlan();
        
        // Send requested  
        if ((mRetrievalStrategy == RETRIEVAL_STRATEGY_ALWAYS)
                || ((mRetrievalStrategy == RETRIEVAL_STRATEGY_NOLOCAL)
                        && (relPlans.isEmpty())))
            sendRequested(anEvent.getMDesire(), intID, "0");
    }

    public void onEventPlanOutcome(CooEventPlanOutcome anEvent) {
        try {
            CooIntention anInt = 
                mIntentions.getIntention(anEvent.getMIntentionID());
            anInt.dispatchPlanOutcome(anEvent);
        }
        catch (CooException e) {
            if (logger.isErrorEnabled()) {
                logger.error("Agent " + mAgentID + " encountered an exception" +
                        " processing a <planOutcome> message for intention" +
                        anEvent.getMIntentionID(), e);
            }
        }
    }

    /**
     * Callback for incoming "provided" events
     * 
     * @param anEvent
     */
    public void onEventProvided(CooEventProvided anEvent) {
        if (logger.isDebugEnabled()) {
            logger.debug("onEventProvided() - Agent " + mAgentID + 
                    ": Processing " + anEvent);
        }


        if (mAcquisitionStrategy == ACQUISITION_STRATEGY_ADD) {
            mPlans.addPlans(anEvent.getMProvidedPlans());
        } else if (mAcquisitionStrategy == ACQUISITION_STRATEGY_REPLACE)
            mPlans.replace(anEvent.getMProvidedPlans());
        else if (mAcquisitionStrategy == ACQUISITION_STRATEGY_DISCARD)
                   ; // nothing to do
        else
            if (logger.isErrorEnabled()) {
                logger.error("onEventProvided() - Agent " + mAgentID + 
                        "has an unknown acquisition policy!", null);
            }
     
        try {
            CooIntention anInt = 
                mIntentions.getIntention(anEvent.getMReqID().getMIntentionID());
            anInt.dispatchProvided(anEvent);
        }
        catch (CooException e) {
            if (logger.isErrorEnabled()) {
                logger.error("Agent " + mAgentID + " encountered an exception" +
                        " processing a <planOutcome> message for intention" +
                        anEvent.getMReqID().getMIntentionID(), e);
            }
        }
        

        // TODO update the intention stack

    }

    /**
     * Callback for incoming "requested" events
     * 
     * @param anEvent
     */
    public void onEventRequested(CooEventRequested anEvent) {
        if (logger.isDebugEnabled()) {
            logger.debug("onEventRequested() - Agent " + mAgentID + 
                    ": Processing " + anEvent);
        }
     
        CooSetOfPlans res = mPlans.getRelevantForAgent(anEvent.getMAgentFROM(),
                anEvent.getMDes());

        if (res.isEmpty()) {
            if (logger.isDebugEnabled()) {
                logger.debug("onEventRequested() - no relevant plans " +
                        "found for " + anEvent.getMDes());
            }

            return;
        }
        else {
            // let's build the reply event...
            CooEventProvided evt = new CooEventProvided(mAgentID, anEvent
                    .getMAgentFROM(), anEvent.getMReqID(), res);
            // ... and send it!
            CooEngine.getAgentsDirectory().sendEvent(evt);
        }
    }

    public void prepare4shutdown() {
        this.mPrepare4shutdown = true;
    }

    public void run() {
        while (!mPrepare4shutdown) {
            try {
                sleep(5000);
            }
            catch (InterruptedException e) {}
        }
        
        //TODO Gestire intenzioni in esecuzione
        
        while (!mShutdown) {
        
        }
        
        // TODO 


        if (logger.isDebugEnabled()) {
            logger.debug("Agent " + mAgentID + " stopped!");
        }
    }

    public void shutdown() {
        this.mShutdown = true;
    }

    /**
     * Creates and returns a snapshot of the agent state
     * 
     * @return
     */
    public CooAgentState getAgentState() {
        return new CooAgentState(mAgentID, mPlans, mPartners,
                mRetrievalStrategy, mAcquisitionStrategy);
    }

    /**
     * 
     * @uml.property name="mEventQueue"
     */
    public CooAgentEventsQueue getMEventQueue() {
        return mEventQueue;
    }

    /**
     * 
     * @uml.property name="mAgentID"
     */
    public String getMAgentID() {
        return mAgentID;
    }
    
    /**
     * Construct and send a "requested" event to the set of partner agents
     * 
     * @param aDes
     *            The desire
     * @param intD
     *            The intention source of the request
     * @param stackPos
     *            The position inside the intention stack of the
     * @param instanceID
     *            The id of the instance record
     */
    private void sendRequested(
            ICooDesire aDes,
            String intID,
            String instanceID) {
        
        CooEventRequested aReq = new CooEventRequested(mAgentID, "",
                new CooRequestID(intID, instanceID), aDes);
        CooEngine.getAgentsDirectory().sendEvent(aReq, mPartners);
    }
}