package org.coows.engine;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.client.Call;
import org.apache.axis.client.Service;
import org.apache.axis.encoding.ser.BeanDeserializerFactory;
import org.apache.axis.encoding.ser.BeanSerializerFactory;

import org.coows.CooAccessSpecifierOnlyTrusted;
import org.coows.CooAccessSpecifierPrivate;
import org.coows.CooAccessSpecifierPublic;
import org.coows.CooBpelProcessBpr;
import org.coows.CooDesire;
import org.coows.CooEventProvided;
import org.coows.CooEventRequested;
import org.coows.CooPlan;
import org.coows.CooEventRemote;
import org.coows.ICooEvent;

import org.coows.ser.CooBprFileDeserializerFactory;
import org.coows.ser.CooBrpFileSerializerFactory;

import org.coows.util.CooBprFile;
import org.coows.util.CooSetOfPlans;

import org.uddi4j.UDDIException;

import org.uddi4j.client.UDDIProxy;

import org.uddi4j.datatype.Name;
import org.uddi4j.datatype.business.BusinessEntity;

import org.uddi4j.response.AuthToken;
import org.uddi4j.response.BusinessDetail;
import org.uddi4j.response.BusinessInfo;
import org.uddi4j.response.BusinessList;

import org.uddi4j.transport.TransportException;
import org.uddi4j.transport.TransportFactory;

import org.uddi4j.util.DiscoveryURL;
import org.uddi4j.util.DiscoveryURLs;
import org.uddi4j.util.FindQualifier;
import org.uddi4j.util.FindQualifiers;

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.rpc.ParameterMode;
import javax.xml.rpc.ServiceException;

/**
 * Interface the CooWS engine with the UDDI agents repository
 */
public class CooAgentsDirectory {
    /**
     * Logger for this class
     */
    private static final Log logger = 
        LogFactory.getLog(CooAgentsDirectory.class);

    private final static QName REMOTEEVENT_QNAME =
        new QName("urn:coows", "RemoteEvent");

    /**
     * A cache for the UDDI repository
     * 
     */
    private static Hashtable<String,String> mUDDIRepositoryCache;

    /**
     * The local hostname
     */
    private final String mHostname;

    /**
     * The UDDI Proxy (see the juddi doc)
     */
    private UDDIProxy mUDDIProxy;

    /**
     * The Auth token (see the juddi doc)
     */
    private AuthToken mAuthToken;

    /**
     * Default constructor
     */
    public CooAgentsDirectory(CooEngineConfiguration aConfig)
            throws CooException {
        try {
            mHostname = InetAddress.getLocalHost().getHostName();
        }
        catch (UnknownHostException e) {
            throw new CooException("[CooAgentsDirectory] init failed! "
                    + "Unable to resolv hostname!");
        }

        initUDDI(aConfig);
        mUDDIRepositoryCache = new Hashtable<String,String>();

        if (logger.isDebugEnabled()) {
            logger.debug("CooAgentsDirectory() - Agents Directory Loaded!");
        }
    }

    /**
     * Find an agent location through the UDDI registry
     * 
     * @param anAgentID
     *            the agentID
     * 
     * @return the host where the agent is hosted
     * 
     * @throws TransportException
     * @throws UDDIException
     */
    private String getAgentLocation(String anAgentID)
            throws UDDIException, TransportException {
        String agentHostedBy;

        if (mUDDIRepositoryCache.containsKey(anAgentID)) {
            // retrieve the agent info from the local cache
            agentHostedBy = mUDDIRepositoryCache.get(anAgentID);
        }
        else {
            // Retrieve the agent info from the remote UDDI registry
            // creating vector of Name Object
            Vector<Name> names = new Vector<Name>();
            names.add(new Name(anAgentID));

            // Setting FindQualifiers to 'exactNameMatch'
            FindQualifiers findQualifiers = new FindQualifiers();
            Vector qualifier = new Vector();
            qualifier.add(new FindQualifier("exactNameMatch"));
            findQualifiers.setFindQualifierVector(qualifier);

            // Find the agent location
            BusinessList businessList = mUDDIProxy.find_business(names, null,
                    null, null, null, findQualifiers, 2);

            Vector businessInfoVector = 
                businessList.getBusinessInfos().getBusinessInfoVector();

            if (businessInfoVector.size() > 1) {
                logger.warn("agent " + anAgentID + 
                        " is published twice!", null);
            }

            // Retrieve the agent's info
            BusinessInfo bi = (BusinessInfo) businessInfoVector.elementAt(0);
            BusinessDetail bd = 
                mUDDIProxy.get_businessDetail(bi.getBusinessKey());
            BusinessEntity be = (BusinessEntity)
                bd.getBusinessEntityVector().elementAt(0);
            agentHostedBy = be.getDiscoveryURLs().get(0).getText();

            // Update the local cache for future lookups
            mUDDIRepositoryCache.put(anAgentID, agentHostedBy);
        }

        if (logger.isDebugEnabled()) {
            logger.debug("getAgentLocation() - retrieved agent's location : "
                    + "agentHostedBy = " + agentHostedBy);
        }

        return agentHostedBy;

    }

    /**
     * Publish the local agents to the UDDI registry
     * 
     * @param config
     * @param enumeration
     * 
     * @throws CooException
     */
    public void publishLocalAgents(Set<String> agents)
            throws CooException {
        publishAgents2UDDI(agents);
    }

    /**
     * Send an event to an agent
     * 
     * @param mAgents
     * @param anEvent
     */
    public void sendEvent(ICooEvent anEvent) {
        if (logger.isDebugEnabled()) {
            logger.debug("sendEvent() - Dispatching event: " + anEvent);
        }

        Dispatcher d = new Dispatcher(anEvent);
    }

    /**
     * Send the same event to different agents
     * 
     * @param anEvent
     * @param dests
     */
    public void sendEvent(CooEventRequested anEvent, Vector dests) {
        if (logger.isDebugEnabled()) {
            logger.debug("sendEvent() - Dispatching event " + anEvent
                    + " to agents " + dests);
        }

        Dispatcher d = new Dispatcher(anEvent, dests);
    }

    /**
     * Unregister agents from the UDDI server
     * 
     * @param agents
     */
    public void unregisterAgentsFromUDDI(Set agents) {
    // TODO
    }

    /**
     * Shutdown the Agents Directory
     * 
     */
    public void shutdown() {
        unregisterAgentsFromUDDI(CooEngine.getMAgentsManager().getAgentsIDs());
    }

    /**
     * Connect to the UDDI server and get an auth token
     * 
     * @throws CooException
     */
    private void initUDDI(CooEngineConfiguration aConfig)
            throws CooException {
        try {
            String aUDDInquiryURL = aConfig
                    .getText("/config/agentsDirConf/uddi-registry/inquiry-url");
            String aUDDIpublishURL = aConfig
                    .getText("/config/agentsDirConf/uddi-registry/publish-url");

            System.setProperty(TransportFactory.PROPERTY_NAME,
                    "org.uddi4j.transport.ApacheAxisTransport");
            System.setProperty("org.uddi4j.logEnabled", "false");
            System.setProperty("java.protocol.handler.pkgs",
                    "com.sun.net.ssl.internal.www.protocol");

            java.security.Security.addProvider((java.security.Provider) Class
                    .forName("com.sun.net.ssl.internal.ssl.Provider")
                    .newInstance());

            mUDDIProxy = new UDDIProxy();
            mUDDIProxy.setInquiryURL(aUDDInquiryURL);
            mUDDIProxy.setPublishURL(aUDDIpublishURL);

            mAuthToken = mUDDIProxy.get_authToken(
              aConfig.getText("/config/agentsDirConf/uddi-registry/userid"),
              aConfig.getText("/config/agentsDirConf/uddi-registry/password"));
        }
        catch (MalformedURLException e) {
            throw new CooException("[CooAgentsDirectory] init failed! "
                    + "UDDI server URL malformed! check the config file!");
        }
        catch (UDDIException e) {
            throw new CooException("[CooAgentsDirectory] init failed! "
                    + "Unable to get an authorization token from "
                    + "the remote UDDI registry " + e);
        }
        catch (TransportException e) {
            throw new CooException("[CooAgentsDirectory] init failed! "
                    + "Unable to get an authorization token from "
                    + "the remote UDDI registry " + e);
        }
        catch (ClassNotFoundException e) {
            throw new CooException("[CooAgentsDirectory] init failed! "
                    + "Error configuring JSSE provider. "
                    + "Make sure JSSE is in classpath! " + e);
        }
        catch (InstantiationException e) {
            throw new CooException("[CooAgentsDirectory] init failed! "
                    + "Error configuring JSSE provider! " + e);
        }
        catch (IllegalAccessException e) {
            throw new CooException("[CooAgentsDirectory] init failed! "
                    + "Error configuring JSSE provider! " + e);
        }
    }

    /**
     * Publish the local agents to the UDDI server
     * 
     * @param agents
     * @throws CooException
     */
    private void publishAgents2UDDI(Set<String> agents)
            throws CooException {
        Vector<BusinessEntity> localAgents = new Vector<BusinessEntity>();
        BusinessEntity be;
        String agentID;
        DiscoveryURLs dus = new DiscoveryURLs();
        dus.add(new DiscoveryURL(mHostname, "hostedBy"));

        for (Iterator<String> i = agents.iterator(); i.hasNext();) {
            agentID = (String) i.next();
            be = new BusinessEntity("", agentID);
            be.setDiscoveryURLs(dus);
            localAgents.add(be);

            // Add the local agents to the UDDI repository cache
            mUDDIRepositoryCache.put(agentID, "localhost");
        }

        try {
            BusinessDetail bd = mUDDIProxy.save_business(mAuthToken
                    .getAuthInfoString(), localAgents);
            System.out.println("[CooAgentsDirectory] "
                    + "Local agents registered to UDDI registry!");
        }
        catch (UDDIException e) {
            throw new CooException(
                    "[CooAgentsDirectory] Error (UDDIException) "
                            + "publishing local agents to UDDI " + e);
        }
        catch (TransportException e) {
            throw new CooException(
                    "[CooAgentsDirectory] Error (TransportException) "
                            + "publishing local agents to UDDI " + e);
        }
    }

    /**
     * Keep the cache updated
     * 
     * @param anAgentID
     * @param anHostname
     */
    public void updateCache(String anAgentID, String anHostname) {
        mUDDIRepositoryCache.put(anAgentID, anHostname);
    }

/**
 * A dispatcher (beta version, use the common-pool!)
 */
    class Dispatcher extends Thread {
        /**
         * Logger for this class
         */
        private final Log logger = LogFactory.getLog(Dispatcher.class);

        Vector<String> mAgents;

        /**
         * 
         * @uml.property name="mEvent"
         * @uml.associationEnd multiplicity="(0 1)"
         */
        ICooEvent mEvent;

        public Dispatcher(ICooEvent e) {
            super(e.toString());
            mAgents = new Vector<String>();
            mAgents.add(e.getMAgentTO());
            mEvent = e;
            start();
        }

        public Dispatcher(ICooEvent e, Vector<String> agents) {
            super(e.toString());
            mAgents = agents;
            mEvent = e;
            start();
        }

        public void run() {

            String curr;

            while (!mAgents.isEmpty()) {
                curr = mAgents.remove(0);
                mEvent.setMAgentTO(curr);
                try {
                    String hostDest = getAgentLocation(curr);

                    
                    // if (hostDest.equals("localhost"))
                    if (!hostDest.equals("localhost"))
                        // TODO RIATTIVARE GUARDIA CORRETTA!!!!
                        // CON QUESTA GUARDIA SPEDISCO
                        // TRAMITE WS AGLI
                        // AGENTI LOCALI
                        CooEngine.getMAgentsManager().dispatchEvent(mEvent);
                    else {
                        Call call = createCall(hostDest);

                        call.invoke(new Object[] { 
                                new CooEventRemote(mEvent, mHostname) });
                    }
                }
                catch (Exception e) {
                    System.out.println("[CooAgentsDirectory]"
                            + "Error sending event to " + curr + ": "
                            + e.getMessage());
                }
            }
        }

        /**
         * Build the Call
         * 
         * @param hostDest
         * 
         * @return
         * 
         * @throws ServiceException
         * @throws MalformedURLException
         */
        private Call createCall(String hostDest)
                throws ServiceException, MalformedURLException {
            Call call = (Call) new Service().createCall();
            call.setTargetEndpointAddress(new URL("http://" + hostDest
                    + ":4444/active-bpel/services/CooWS"));

            // Registering type mapping
            call.registerTypeMapping(CooDesire.class, new QName("urn:coows",
                    "Desire"), BeanSerializerFactory.class,
                    BeanDeserializerFactory.class);
            call.registerTypeMapping(CooPlan.class, new QName("urn:coows",
                    "Plan"), BeanSerializerFactory.class,
                    BeanDeserializerFactory.class);
            call.registerTypeMapping(CooAccessSpecifierPublic.class, new QName(
                    "urn:coows", "AccessSpecifierPublic"),
                    BeanSerializerFactory.class, BeanDeserializerFactory.class);
            call.registerTypeMapping(CooAccessSpecifierPrivate.class,
                    new QName("urn:coows", "AccessSpecifierPrivate"),
                    BeanSerializerFactory.class, BeanDeserializerFactory.class);
            call.registerTypeMapping(CooAccessSpecifierOnlyTrusted.class,
                    new QName("urn:coows", "AccessSpecifierOnlyTrusted"),
                    BeanSerializerFactory.class, BeanDeserializerFactory.class);
            call.registerTypeMapping(CooSetOfPlans.class, new QName(
                    "urn:coows", "SetOfPlans"), BeanSerializerFactory.class,
                    BeanDeserializerFactory.class);
            call.registerTypeMapping(CooBpelProcessBpr.class, new QName(
                    "urn:coows", "BpelProcessBpr"),
                    BeanSerializerFactory.class, BeanDeserializerFactory.class);
            call.registerTypeMapping(CooBprFile.class, new QName("urn:coows",
                    "BprFile"), CooBrpFileSerializerFactory.class,
                    CooBprFileDeserializerFactory.class);
            call.registerTypeMapping(ICooEvent.class, new QName("urn:coows",
                    "Event"), BeanSerializerFactory.class,
                    BeanDeserializerFactory.class);
            call.registerTypeMapping(CooEventRequested.class, new QName(
                    "urn:coows", "EventRequested"),
                    BeanSerializerFactory.class, BeanDeserializerFactory.class);
            call.registerTypeMapping(CooEventProvided.class, new QName(
                    "urn:coows", "EventProvided"), BeanSerializerFactory.class,
                    BeanDeserializerFactory.class);
            call.registerTypeMapping(CooEventRemote.class, new QName(
                    "urn:coows", "RemoteEvent"), BeanSerializerFactory.class,
                    BeanDeserializerFactory.class);

            call.setOperationStyle("rpc");
            call.setOperationName(new QName("CooWS", "dispatchEventRemote"));
            call.addParameter(REMOTEEVENT_QNAME, REMOTEEVENT_QNAME,
                    CooEventRemote.class, ParameterMode.IN);

            return call;
        }
    }

}
