package org.coows.engine;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.coows.CooAgent;
import org.coows.CooEventProvided;
import org.coows.CooEventRemote;
import org.coows.CooEventRequested;
import org.coows.CooGlobalEventsQueue;
import org.coows.ICooEvent;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

/**
 * The manager of local agents
 */
public class CooAgentsManager {
    /**
     * Logger for this class
     */
    private static final Log logger = LogFactory.getLog(CooAgentsManager.class);

    /**
     * A table holding the ref to local agents
     */
    private Hashtable<String,CooAgent> mAgents;

    /**
     * The queue where are stored the incoming events waiting to be dispatched
     * 
     * @uml.property name="mGlobalEventsQueue"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private CooGlobalEventsQueue mGlobalEventsQueue;


    /**
     * Restore the agents with the last saved state and initialize all local
     * variables (thread pool and global event queue)
     * 
     * @param aConf
     * @throws CooException
     */
    public CooAgentsManager(CooEngineConfiguration aConf)
        throws CooException {
        if (logger.isDebugEnabled()) {
            logger.debug("CooAgentsManager() - "
                    + "initializing Agents Manager structures");
        }

        // Restore the agents
        mAgents = CooPersistenceManager.restoreAgents(aConf);

        // Create the global events queue
        mGlobalEventsQueue = new CooGlobalEventsQueue();

        if (logger.isDebugEnabled()) {
            logger.debug("CooAgentsManager() - Agents Manager loaded!");
        }
    }
    
    
    /**
     * Start the agents execution
     * 
     */
    public void startAgents() {
        for (Iterator<CooAgent> i = mAgents.values().iterator();i.hasNext();) {
            i.next().start();
        }
    }
    



    /**
     * Handle incoming events throuth the global events queue <br>
     * 
     * 
     * @param anEvent
     * @throws CooException
     */
    public void dispatchEvent(ICooEvent anEvent) {
        if (logger.isDebugEnabled()) {
            logger.debug("dispatchEvent(anEvent = " + anEvent
                    + ") - received new incoming event");
        }

        mGlobalEventsQueue.enqueue(anEvent);
    }
    
    /**
     * Dispatch an incoming remote event <br>
     * On event requested the agents directory cache is updated <br>
     * On event provided the processes manager is notified on the source host
     * 
     * @param anEvent
     * @throws CooException
     */
    public void dispatchEvent(CooEventRemote anEvent) {
        if (anEvent.getMEvent() instanceof CooEventRequested)
            CooEngine.getAgentsDirectory().updateCache(
                    ((CooEventRequested) anEvent.getMEvent()).getMAgentFROM(),
                    anEvent.getMSourceHost());
        else
            if (anEvent.getMEvent() instanceof CooEventProvided)
                CooEngine.getMProcessesManager().updateNoLocal(
                        ((CooEventProvided) anEvent.getMEvent())
                                .getMProvidedPlans(), anEvent.getMSourceHost());

        this.dispatchEvent(anEvent.getMEvent());
    }

    /**
     * Shutdown the agents manager<br>
     * Serialize the current agents states to the db and halt the agents 
     * execution cicle
     *
     * @param aConf
     */
    public void shutdown(CooEngineConfiguration aConf) {

        Iterator<CooAgent> curr;
        
        try {
            // Clear the global queue
            mGlobalEventsQueue.clear();
            
            // Notify the agents
            for (curr = mAgents.values().iterator();curr.hasNext();) 
                curr.next().prepare4shutdown();
            
            
            // Serialize agents state to the db
            CooPersistenceManager.serializeAgents(aConf, mAgents.values());
    
            // Halt the agents execution
            for (curr = mAgents.values().iterator(); curr.hasNext();) {
                curr.next().shutdown();
            }
    
            if (logger.isDebugEnabled()) {
                logger.debug("shutdown() done!");
            }
        } catch (CooException e) {
            if (logger.isErrorEnabled()) {
                logger.error("shutdown() - Encountered an error while" +
                        "shuting down the Agents Manager!",e);
            }
        }
        
    }


    /**
     * getter
     * @param anAgentID
     * @return
     */
    public CooAgent getAgent(String anAgentID) {
        return (CooAgent) mAgents.get(anAgentID);
    }

    public Set<String> getAgentsIDs() {
        return mAgents.keySet();
    }

    /**
     * getter
     * @uml.property name="mGlobalEventsQueue"
     */
    public CooGlobalEventsQueue getMGlobalEventsQueue() {
        return mGlobalEventsQueue;
    }

}
