package org.coows.engine;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.coows.CooBpelProcessParameters;
import org.coows.CooEventPlanOutcome;
import org.coows.CooPlan;
import org.coows.ICooBpelProcess;
import org.coows.util.CooSetOfPlans;

import java.net.InetAddress;
import java.net.UnknownHostException;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

/**
 * The Manager of the local BPEL Processes
 * 
 */
public class CooBpelProcessesManager {
    /**
     * Logger for this class
     */
    private static final Log logger = 
        LogFactory.getLog(CooBpelProcessesManager.class);

    /**
     * The archive of local BPEL processes The table maps for each serviceName
     * the corresponding BPEL process
     */
    private Hashtable<String,ICooBpelProcess> mProcesses;

    /**
     * This table holds the info necessary for the retrieval of remote processes
     * For each serviceName we have a list of hostnames wich coud supply the
     * BPEL process associated
     */
    private Hashtable<String,Vector<String>> mNoLocalProcesses;

    /**
     * The local hostname
     */
    private String mHostname;

    /**
     * List of serviceName currently deployed
     */
    private Vector<String> mProcessesDeployed;

    /**
     * Constructor
     * 
     * @param aConf
     *            The engine configuration
     * @throws CooException
     */
    public CooBpelProcessesManager(CooEngineConfiguration aConf)
            throws CooException {
        mProcesses = CooPersistenceManager.restoreBpelProcesses(aConf);
        mProcessesDeployed = new Vector<String>();
        mNoLocalProcesses = new Hashtable<String,Vector<String>>();

        try {
            // retrieve the hostname
            mHostname = InetAddress.getLocalHost().getHostName();
        }
        catch (UnknownHostException e) {
            mHostname = "localhost";
            System.out.println("[CooBpelProcessManager.init] WARNING: "
                    + "unable to resolv hostname! Using localhost");
        }

        if (logger.isInfoEnabled()) {
            logger.info("CooBpelProcessesManager() - "
                    + "BPEL Processes Manager Loaded!");
        }
    }

    /**
     * Invoke a process on demand
     * 
     * @param anAgentID
     *            The agent requestor
     * @param aServiceName
     *            The service name
     */
    public void invokeProcess(String anAgentID, String anIntentionID,
            String aServiceName) {

        if (logger.isDebugEnabled()) {
            logger.debug("runProcess(agentID = " + anAgentID + ", "
                    + "serviceName = " + aServiceName + ")");
        }

        // If the process is not stored locally and doesn't appear
        // among the remote provided processes
        if (!mProcesses.containsKey(aServiceName)
                && !mNoLocalProcesses.containsKey(aServiceName)) {
            
            if (logger.isErrorEnabled()) {
                logger.error("runProcess() Error! " +
                  "Trying to run unknown process!!! " + aServiceName, null);
            }

            return;
        }

        // If the process is stored on a remote node
        if (!mProcesses.containsKey(aServiceName)
                && mNoLocalProcesses.contains(aServiceName)) {
            // Retrieve the remote BPEL process
            String remoteHost = 
                mNoLocalProcesses.get(aServiceName).lastElement();
            // TODO send to the remote host the request
        }

        ICooBpelProcess proc = (ICooBpelProcess) mProcesses.get(aServiceName);

        try {
            if (!mProcessesDeployed.contains(aServiceName)) {
                // Deploy the process
                proc.deploy();
                mProcessesDeployed.add(aServiceName);
            }

            // Invoke the process
            proc.invoke(
             new CooBpelProcessParameters(mHostname, anAgentID,anIntentionID));
        } catch (CooException e) {
            logger.error("runProcess(serviceName = " + aServiceName
                    + ") - Failed to invoke process ", e);

            // Notify the agent requester
            CooEngine.getMAgentsManager().dispatchEvent(
                    new CooEventPlanOutcome(anAgentID,anIntentionID,false));
        }
    }

    /**
     * 
     * @param aServiceName
     * @return
     * @throws CooException
     */
    public ICooBpelProcess getProcess(String aServiceName)
            throws CooException {
        if (mProcesses.containsKey(aServiceName))
            return (ICooBpelProcess) mProcesses.get(aServiceName);
        else
            throw new CooException("Process " + aServiceName + "not found!");
    }

    /**
     * Keep trace of the host where could be retrieved the BPEL processes<br>
     * of the plans retrieved remotely
     * 
     * @param aSop
     * @param anHostname
     */
    public void updateNoLocal(CooSetOfPlans aSop, String anHostname) {
        CooPlan curr;
        Vector<String> aux;

        for (Iterator<CooPlan> i = aSop.getMPlans().iterator(); i.hasNext();) {
            curr = i.next();

            if (mProcesses.containsKey(curr.getMPlanID()))
                continue;
            else {
                aux = (Vector<String>) mNoLocalProcesses.get(curr.getMPlanID());
                if (null == aux) aux = new Vector<String>();
                aux.add(anHostname);

                mNoLocalProcesses.put(curr.getMPlanID(), aux);
            }
        }
    }

    /**
     * Shutdown the Manager <br>
     * (undeploys all the processes and stores them back to the db)
     * 
     * @param aConf
     * @throws CooException
     */
    public void shutdown(CooEngineConfiguration aConf)
            throws CooException {
        
        for (Iterator<String> i = mProcessesDeployed.iterator(); i.hasNext();)
            mProcesses.get(i.next()).undeploy();
        
        CooPersistenceManager.serializeProcesses(aConf, mProcesses.values());

        if (logger.isDebugEnabled()) {
            logger.debug("shutdown() done!");
        }
    }

    /**
     * Dispatch an <b>achieved</b> message to a suspended process 
     * @param planID
     */
    public void dispatchAchieved(String anAgentID, String anIntentionID,
            String aServiceName) {
        
    }
}
