package org.coows.engine;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Singleton class
 */
public class CooEngine {
    /**
     * Logger for this class
     */
    private static final Log logger = LogFactory.getLog(CooEngine.class);
    

	public static final String CATALINA_HOME =
        System.getProperties().getProperty("catalina.home");

	public static final String DEFAULT_CONFIG_DIR = CATALINA_HOME + "/conf";

	private static final String DEFAULT_CONFIG_FILE = "cooEngineConf.xml";

    /**
     * The in memory representation of the engine configuration
     * 
     * @uml.property name="mConfig"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private static CooEngineConfiguration mConfig;

    /**
     * The agents manager
     * 
     * @uml.property name="mAgentsManager"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private static CooAgentsManager mAgentsManager;

    /**
     * The agents directory
     * 
     * @uml.property name="mAgentsDirectory"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private static CooAgentsDirectory mAgentsDirectory;

    /**
     * The manager of the bpel processes
     * 
     * @uml.property name="mProcessesManager"
     * @uml.associationEnd multiplicity="(0 1)"
     */
    private static CooBpelProcessesManager mProcessesManager;



	public static CooAgentsDirectory getAgentsDirectory() {
		return mAgentsDirectory;
	}

    /**
     * 
     * @uml.property name="mAgentsManager"
     */
    public static CooAgentsManager getMAgentsManager() {
        return mAgentsManager;
    }

    /**
     * 
     * @uml.property name="mConfig"
     */
    public static CooEngineConfiguration getMConfig() {
        return mConfig;
    }



	public static void init() {
		init(DEFAULT_CONFIG_FILE, DEFAULT_CONFIG_DIR);
	}


    // Convenience method for debug purpose
	public static void initTEMP(String configFile, String configDir) {
        try {
			mConfig = new CooEngineConfiguration(configFile, configDir);
		} catch (Exception e) {
		}
	}


    /**
     * Initialize all the CooWS modules and start the execution
     * 
     * @param configFile
     * @param configDir
     */
	public static void init(String configFile, String configDir) {
        if (logger.isDebugEnabled())
        {
            logger.debug("init() - Initializing CooWS engine components");
        }

		try {
            // Load the configuration file
			mConfig = new CooEngineConfiguration(configFile, configDir);

			// PopulateDB.populateProcessTable(mConfig);
            
            // Initialize the BPEL Processes Manager 
			mProcessesManager = new CooBpelProcessesManager(mConfig);

            // Initialize the Agents Manager
			mAgentsManager = new CooAgentsManager(mConfig);

			// CooPersistenceManager.restoreAgents(mConfig);
            
            // Initialize the Agents Directory Manager
			mAgentsDirectory = new CooAgentsDirectory(mConfig);
            
            // Update the directory with the local agents
			mAgentsDirectory.publishLocalAgents(mAgentsManager.getAgentsIDs());

            // Start the agents execution
			mAgentsManager.startAgents();
            
            if (logger.isDebugEnabled()) {
                logger.debug("init() - CooWS engine started!");
            }
		} catch (Exception e) {
            if (logger.isErrorEnabled())
            {
                logger.error("init() - Failed to start the CooWS engine!", e);
            }
            
		}
	}


    /**
     * Shutdown the agents 
     * @throws CooException
     */
	public static void shutdown() throws CooException {
        if (logger.isDebugEnabled())
        {
            logger.debug("shutdown() - Halting the CooWS engine");
        }
        

		mAgentsDirectory.shutdown();
		mAgentsManager.shutdown(mConfig);
		mProcessesManager.shutdown(mConfig);
        
        
        if (logger.isDebugEnabled())
        {
            logger.debug("shutdown() - " +
                    "CooWS engine has been successfully shutdown!");
        }
	}

    /**
     * 
     * @uml.property name="mProcessesManager"
     */
    public static CooBpelProcessesManager getMProcessesManager() {
        return mProcessesManager;
    }

}
