package org.coows.engine;

/**
 * Base class for all the CooWS exceptions
 */
public class CooException extends Exception {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/** Flag indicating if informational logging is enabled, off by default. */
	public static boolean sInfoIsOn;

	/** Flag indicating if stack trace enabled, off by default. */
	public static boolean sStackTraceOn = true;

	/** The root cause of this exception */
	protected Throwable mRootCause;


	/**
	 * Construct a new runtime exception.
	 */
	public CooException() {
		super();
	}


	/**
	 * Construct a new runtime exception with the passed info string.
	 * 
	 * @see java.lang.Throwable#Throwable(String)
	 */
	public CooException(String aInfo) {
		super(aInfo);
	}


	/**
	 * Construct with a root exception (used primarily for rethrowing an
	 * underlying exception).
	 * 
	 * @param aRootCause
	 */
	public CooException(Throwable aRootCause) {
		super(aRootCause);
		setRootCause(aRootCause);
	}


	/**
	 * Construct with a root exception (used prinarily for rethrowing an
	 * underlying exception).
	 * 
	 * @param aInfo
	 *            Informational message for the exception
	 * @param aRootCause
	 *            Root cause of the exception
	 */
	public CooException(String aInfo, Throwable aRootCause) {
		super(aInfo, aRootCause);
		setRootCause(aRootCause);
	}


	/**
	 * Sets informational logging enabled or disabled based on flag.
	 * 
	 * @param aFlag
	 *            True enables logging and False disables logging
	 */
	public static void setInfoOn(boolean aFlag) {
		sInfoIsOn = aFlag;
	}


	/**
	 * Sets the exception which is the Root Cause of this exception.
	 * 
	 * @param aRootCause
	 *            The Root Cause to set
	 */
	public void setRootCause(Throwable aRootCause) {
		mRootCause = aRootCause;
	}


	/**
	 * Returns the exception which is the Root Cause of this exception.
	 * 
	 * @return Exception causing this exception, null if none.
	 */
	public Throwable getRootCause() {
		return mRootCause;
	}


	/**
	 * Sets stack trace logging enabled or disabled based on flag.
	 * 
	 * @param aFlag
	 *            True enables logging and False disables stack trace logging
	 */
	public static void setStackTraceOn(boolean aFlag) {
		sStackTraceOn = aFlag;
	}


	/**
	 * Utility method to write info to the console.
	 * 
	 * @param aInfo
	 *            Informational message to be displayed.
	 */
	public static void info(String aInfo) {
		if (infoOn()) {
			System.out.println(aInfo);
		}
	}


	/**
	 * Returns true if info output is enabled.
	 */
	public static boolean infoOn() {
		return sInfoIsOn;
	}


	/**
	 * Utility method to write info to the console.
	 * 
	 * @param aPrompt
	 *            Informational message to be displayed.
	 */
	public static void logError(Throwable aThrowable, String aSummary) {
		System.out.println(aSummary);

		if (sStackTraceOn && (aThrowable != null)) {
			aThrowable.printStackTrace();
		}
	}


	/**
	 * Returns true if stack trace output is enabled.
	 */
	public static boolean stackTraceOn() {
		return sInfoIsOn;
	}


	/**
	 * Logs this exception error to the console.
	 */
	public void logError() {
		System.out.println(getLocalizedMessage());

		if (getRootCause() == null) {
			if (stackTraceOn()) {
				printStackTrace();
			}
		} else {
			System.out.print("Caused by: ");

			if (getRootCause() instanceof CooException) {
				((CooException) getRootCause()).logError();
			} else {
				System.out.println(getRootCause().getLocalizedMessage());

				if (stackTraceOn()) {
					getRootCause().printStackTrace();
				}
			}
		}
	}
}
